import React from "react";
import { ReactNode } from "react";
import { Node } from "../structures/graph/node";
import "./selectlayer.css";

type propTypes = {
    children: any;
    allNodes: Node[];
    isEnable: () => boolean;
    screen2GraphCoords: (x: number, y: number) => any;
    onBoxSelect: (nodes: Node[]) => void;
};

type layerCoordinates = {
    x: number;
    y: number;
};

export class SelectLayer extends React.Component<propTypes> {
    private layerContainer: any;
    private layerBox: any;
    private initialSelectPoint: layerCoordinates = undefined;

    constructor(props: propTypes) {
        super(props);

        this.isSelecting = this.isSelecting.bind(this);
        this.onBoxSelect = this.onBoxSelect.bind(this);
        this.boxSelectOnPointerDown = this.boxSelectOnPointerDown.bind(this);
        this.boxSelectOnPointerMove = this.boxSelectOnPointerMove.bind(this);
        this.boxSelectOnPointerUp = this.boxSelectOnPointerUp.bind(this);

        this.layerContainer = React.createRef();
        this.layerBox = React.createRef();
    }

    componentDidMount(): void {
        this.setupBoxSelect();
    }

    setupBoxSelect() {
        // Source: https://github.com/vasturiano/force-graph/issues/151#issuecomment-735850938
        this.layerContainer.current.onpointerdown = this.boxSelectOnPointerDown;
        this.layerContainer.current.onpointermove = this.boxSelectOnPointerMove;
        this.layerContainer.current.onpointerup = this.boxSelectOnPointerUp;
    }

    private isSelecting(): boolean {
        if (!this.initialSelectPoint) {
            return false;
        }

        if (!this.props.isEnable()) {
            this.initialSelectPoint = undefined;
            this.layerBox.current.className = "";
            return false;
        }

        return true;
    }

    onBoxSelect(left: number, bottom: number, top: number, right: number) {
        // Filter out selected nodes
        const hitNodes: Node[] = [];
        const tl = this.props.screen2GraphCoords(left, top);
        const br = this.props.screen2GraphCoords(right, bottom);
        this.props.allNodes.forEach((node: any) => {
            if (
                tl.x < node.x &&
                node.x < br.x &&
                br.y > node.y &&
                node.y > tl.y
            ) {
                // Add node if in box area
                hitNodes.push(node);
            }
        });

        this.props.onBoxSelect(hitNodes);
    }

    boxSelectOnPointerDown(e: any) {
        if (!this.props.isEnable()) {
            return;
        }

        e.preventDefault();
        this.layerBox.current.style.left = e.offsetX.toString() + "px";
        this.layerBox.current.style.top = e.offsetY.toString() + "px";
        this.layerBox.current.style.width = "0px";
        this.layerBox.current.style.height = "0px";
        this.initialSelectPoint = {
            x: e.offsetX,
            y: e.offsetY,
        };
        this.layerBox.current.className = "visible";
    }

    boxSelectOnPointerMove(e: any) {
        if (!this.isSelecting()) {
            return;
        }

        e.preventDefault();
        if (e.offsetX < this.initialSelectPoint.x) {
            this.layerBox.current.style.left = e.offsetX.toString() + "px";
            this.layerBox.current.style.width =
                (this.initialSelectPoint.x - e.offsetX).toString() + "px";
        } else {
            this.layerBox.current.style.left =
                this.initialSelectPoint.x.toString() + "px";
            this.layerBox.current.style.width =
                (e.offsetX - this.initialSelectPoint.x).toString() + "px";
        }
        if (e.offsetY < this.initialSelectPoint.y) {
            this.layerBox.current.style.top = e.offsetY.toString() + "px";
            this.layerBox.current.style.height =
                (this.initialSelectPoint.y - e.offsetY).toString() + "px";
        } else {
            this.layerBox.current.style.top =
                this.initialSelectPoint.y.toString() + "px";
            this.layerBox.current.style.height =
                (e.offsetY - this.initialSelectPoint.y).toString() + "px";
        }
    }

    boxSelectOnPointerUp(e: any) {
        if (!this.isSelecting()) {
            return;
        }

        e.preventDefault();
        let left, bottom, top, right;
        if (e.offsetX < this.initialSelectPoint.x) {
            left = e.offsetX;
            right = this.initialSelectPoint.x;
        } else {
            left = this.initialSelectPoint.x;
            right = e.offsetX;
        }
        if (e.offsetY < this.initialSelectPoint.y) {
            top = e.offsetY;
            bottom = this.initialSelectPoint.y;
        } else {
            top = this.initialSelectPoint.y;
            bottom = e.offsetY;
        }
        this.initialSelectPoint = undefined;
        this.layerBox.current.className = "";
        this.onBoxSelect(left, bottom, top, right);
    }

    render(): ReactNode {
        return (
            <div ref={this.layerContainer} id="select-layer">
                <div ref={this.layerBox} id="box-select"></div>
                {this.props.children}
            </div>
        );
    }
}
